import numpy as np
from numpy.typing import NDArray
from polars import DataFrame

from ._typealias import _Csx, _CsxArray

def sparse_kde_csx_py(
    counts: _Csx, kernel: NDArray[np.float32], *, threshold: float = 0
) -> _CsxArray:
    """
    Calculate the KDE for each spot with counts as uint16.
    """

def kde_at_coord(
    counts: GridCounts,
    genes: list[str],
    kernel: NDArray[np.float32],
    coordinates: tuple[NDArray[np.int_], NDArray[np.int_]],
    *,
    n_threads: int | None = None,
) -> _CsxArray:
    """
    Calculate KDE at the given coordinates.
    """

def categorical_coordinate(
    x: NDArray[np.int32], y: NDArray[np.int32], *, n_threads: int | None = None
) -> tuple[NDArray[np.int32], NDArray[np.int32]]:
    """
    Get the codes and the coordinates (comparable to a pandas.Categorical)
    """

def coordinate_as_string(
    x: NDArray[np.int32], y: NDArray[np.int32], *, n_threads: int | None = None
) -> NDArray[np.str_]:
    """
    Concatenate two int arrays elementwise into a string representation (i.e. 'x_y').
    """

def cosinef32_and_celltypei8(
    counts: GridCounts,
    genes: list[str],
    signatures: NDArray[np.float32],
    kernel: NDArray[np.float32],
    *,
    log: bool = False,
    chunk_size: tuple[int, int] = (500, 500),
    n_threads: int | None = None,
) -> tuple[NDArray[np.float32], NDArray[np.float32], NDArray[np.int8]]:
    """
    Calculate the cosine similarity given counts and signatures and assign the most
    similar celltype.
    """

def cosinef32_and_celltypei16(
    counts: GridCounts,
    genes: list[str],
    signatures: NDArray[np.float32],
    kernel: NDArray[np.float32],
    *,
    log: bool = False,
    chunk_size: tuple[int, int] = (500, 500),
    n_threads: int | None = None,
) -> tuple[NDArray[np.float32], NDArray[np.float32], NDArray[np.int16]]:
    """
    Calculate the cosine similarity given counts and signatures and assign the most
    similar celltype.
    """

class GridCounts:
    """
    Object holding each gene as count data in a sparse 2D-grid.
    """

    shape: tuple[int, int]
    """
    tuple[int, int]: Shape of the count arrays.
    """

    def __init__(
        self,
        counts: dict[str, _Csx],
        *,
        resolution: float | None = None,
        n_threads: int | None = None,
    ):
        """
        Parameters
        ----------
        counts : dict[str, scipy.sparse.csr_array | scipy.sparse.csr_matrix | scipy.sparse.csc_array | scipy.sparse.csc_matrix]
            Gene counts.
        resolution : float, optional
            Resolution as nm / pixel.
        n_threads : int, optional
            Number of threads used for reading and processing file. If `None` this will
            default to the number of logical CPUs.

        Raises
        ------
        ValueError
            If genes in `counts` do not all have the same shape.
        """

    @classmethod
    def from_dataframe(
        cls,
        df: DataFrame,
        *,
        resolution: float | None = None,
        binsize: float | None = None,
        n_threads: int | None = None,
    ):  # -> Self
        """
        Initialize from dataframe.

        Transform a :py:class:`polars.DataFrame` that provides a 'gene', 'x', and 'y'
        column into :py:class:`sainsc.GridCounts`. If a 'count' column exists it will
        be used as counts else a count of 1 (single molecule) per row will be assumed.

        Parameters
        ----------
        df : polars.DataFrame
            The data to be transformed.
        binsize : float or None, optional
            The size to bin the coordinates by. If None coordinates must be integers.
        resolution : float, optional
            Resolution of each coordinate unit in nm. The default is 1,000 i.e. measurements
            are in um.
        n_threads : int, optional
            Number of threads used for initializing :py:class:`sainsc.LazyKDE`.
            If `None` this will default to the number of logical CPUs.

        Returns
        -------
        sainsc.GridCounts
        """

    def as_dataframe(self) -> DataFrame:
        """
        Convert to a dataframe with 'gene', 'x', 'y', and 'count' column.

        Returns
        -------
        polars.DataFrame
        """

    def __getitem__(self, key: str) -> _CsxArray: ...
    def __setitem__(self, key: str, value: _Csx): ...
    def __delitem__(self, key: str): ...
    def __len__(self) -> int: ...
    def __contains__(self, item: str) -> bool: ...
    def __eq__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...
    def get(self, key: str, default: _CsxArray | None = None) -> _CsxArray | None:
        """
        Get the counts for a gene.

        Parameters
        ----------
        key : str
            Name of the gene to retrieve.

        Returns
        -------
        scipy.sparse.csr_array | scipy.sparse.csc_array | None
        """

    def genes(self) -> list[str]:
        """
        Get all available genes.

        Returns
        -------
        list[str]
        """

    def gene_counts(self) -> dict[str, int]:
        """
        Number of counts per gene.

        Returns
        -------
        dict[str, int]
            Mapping from gene to number of counts.
        """

    def grid_counts(self) -> NDArray[np.uintc]:
        """
        Counts per pixel.

        Aggregates counts across all genes.

        Returns
        -------
        numpy.ndarray[numpy.uintc]
        """

    def select_genes(self, genes: set[str]):
        """
        Keep selected genes.

        Parameters
        ----------
        genes : set[str]
            List of gene names to keep.
        """

    def filter_genes_by_count(self, min: int = 1, max: int = 4_294_967_295):
        """
        Filter genes by minimum and maximum count thresholds.

        Parameters
        ----------
        min : int, optional
            Minimum count threshold.
        max : int, optional
            Maximum count threshold.
        """

    def crop(self, x: tuple[int | None, int | None], y: tuple[int | None, int | None]):
        """
        Crop the field of view for all genes.

        Parameters
        ----------
        x : tuple[int | None, int | None]
            Range to crop as `(xmin, xmax)`
        y : tuple[int | None, int | None]
            Range to crop as `(ymin, ymax)`
        """

    def filter_mask(self, mask: NDArray[np.bool_]):
        """
        Filter all genes with a binary mask.

        Parameters
        ----------
        mask : numpy.ndarray[numpy.bool]
            All counts where `mask` is `False` will be set to 0.
        """

    @property
    def resolution(self) -> float | None:
        """
        float | None: Resolution in nm / pixel.

        Raises
        ------
            TypeError
                If setting with a type other than `float` or `int`.
        """

    @resolution.setter
    def resolution(self, resolution: float): ...
    @property
    def n_threads(self) -> int:
        """
        int: Number of threads used for processing.

        Raises
        ------
            TypeError
                If setting with a type other than `int` or less than 0.
        """

    @n_threads.setter
    def n_threads(self, n_threads: int): ...



from collections.abc import Iterable
from typing import Protocol, TypeVar

import numpy as np
import pandas as pd
from anndata import AnnData
from numba import njit
from numpy.typing import NDArray
from scipy.sparse import csr_matrix, sparray, spmatrix
from skimage.measure import label, regionprops

from .._utils import _get_coordinate_index
from .._utils_rust import GridCounts

T = TypeVar("T", bound=np.number)
U = TypeVar("U", bound=np.bool_ | np.integer)

_SCALEBAR = {"box_alpha": 0, "color": "w"}


@njit
def _apply_color(
    img_in: NDArray[np.integer], cmap: tuple[NDArray[T], ...]
) -> NDArray[T]:
    img = np.empty(shape=(*img_in.shape, 3), dtype=cmap[0].dtype)
    for i in range(img_in.shape[0]):
        for j in range(img_in.shape[1]):
            img[i, j, :] = cmap[img_in[i, j]]
    return img


def _get_cell_dtype(n: int) -> np.dtype:
    return np.result_type("int8", n)


def _filter_blobs(labeled_map: NDArray[U], min_blob_area: int) -> NDArray[U]:
    # remove small blops (i.e. "cells")
    if min_blob_area <= 0:
        raise ValueError("Area must be bigger than 0")
    blob_labels = label(labeled_map, background=0)
    for blop in regionprops(blob_labels):
        if blop.area_filled < min_blob_area:
            min_x, min_y, max_x, max_y = blop.bbox
            labeled_map[min_x:max_x, min_y:max_y][blop.image] = 0
    return labeled_map


def _localmax_anndata(
    kde: spmatrix | sparray | NDArray,
    genelist: Iterable[str],
    coord: tuple[NDArray[np.integer], ...],
    *,
    name: str | None = None,
    n_threads: int = 1,
) -> AnnData:
    obs = pd.DataFrame(
        index=_get_coordinate_index(*coord, name=name, n_threads=n_threads)
    )

    return AnnData(
        X=csr_matrix(kde),
        obs=obs,
        var=pd.DataFrame(index=pd.Index(genelist, name="gene")),
        obsm={"spatial": np.column_stack(coord)},
    )


class CosineCelltypeCallable(Protocol):
    def __call__(
        self,
        counts: GridCounts,
        genes: list[str],
        signatures: NDArray[np.float32],
        kernel: NDArray[np.float32],
        *,
        log: bool = ...,
        chunk_size: tuple[int, int] = ...,
        n_threads: int | None = ...,
    ) -> tuple[NDArray[np.float32], NDArray[np.float32], NDArray[np.signedinteger]]: ...



from pathlib import Path
from typing import TYPE_CHECKING, Literal, get_args

import numpy as np
import pandas as pd
import polars as pl
from anndata import AnnData
from scipy.sparse import csr_matrix

from .._typealias import _PathLike
from .._utils import (
    _bin_coordinates,
    _get_coordinate_index,
    _get_n_cpus,
    _raise_module_load_error,
)
from .._utils_rust import GridCounts
from ..lazykde import LazyKDE
from ._io_utils import _categorical_coordinate, _open_file
from ._stereoseq_chips import CHIP_RESOLUTION

if TYPE_CHECKING:
    from spatialdata import SpatialData


# Stereo-seq
def _get_stereo_header(filepath: _PathLike) -> dict[str, str]:
    header = dict()
    with _open_file(filepath, "rb") as f:
        for line in f:
            assert isinstance(line, bytes)
            if not line.startswith(b"#"):
                break
            key, value = line.decode().strip("#\n").split("=", 1)
            header[key] = value

    return header


def _get_stereo_resolution(name: str) -> int | None:
    for chip_name in CHIP_RESOLUTION.keys():
        if name.startswith(chip_name):
            return CHIP_RESOLUTION[chip_name]


def read_gem_file(
    filepath: _PathLike, *, sep: str = "\t", n_threads: int | None = None, **kwargs
) -> pl.DataFrame:
    """
    Read a GEM file into a DataFrame.

    GEM files are used by e.g. Stereo-Seq and Nova-ST.

    The gene-ID and count column will be renamed to `gene` and `count`, respectively.

    The name of the count column must be one of `MIDCounts`, `MIDCount`, or `UMICount`.

    Parameters
    ----------
    filepath : os.PathLike or str
        Path to the GEM file.
    sep : str, optional
        Separator used in :py:func:`polars.read_csv`.
    n_threads : int, optional
        Number of threads used for reading and processing file. If `None` this will
        default to the number of available CPUs.
    kwargs
        Other keyword arguments will be passed to :py:func:`polars.read_csv`.

    Returns
    -------
    polars.DataFrame

    Raises
    ------
    ValueError
        If count column has an unknown name.
    """
    # according to specification the name should be MIDCount
    # however the datasets published in the original Stereo-seq publication used
    # various names for the count column
    _Count_ColName = Literal["MIDCounts", "MIDCount", "UMICount"]
    options = get_args(_Count_ColName)

    if n_threads is None:
        n_threads = _get_n_cpus()

    df = pl.read_csv(
        Path(filepath),
        separator=sep,
        comment_prefix="#",
        schema_overrides={"geneID": pl.Categorical, "x": pl.Int32, "y": pl.Int32}
        | {n: pl.UInt32 for n in options},
        n_threads=n_threads,
        **kwargs,
    )

    for name in options:
        if name in df.columns:
            return df.rename({name: "count", "geneID": "gene"})

    raise ValueError(
        f"Unknown count column, the name of the column must be one of {options}"
    )


def read_StereoSeq(
    filepath: _PathLike,
    *,
    resolution: float | None = None,
    sep: str = "\t",
    n_threads: int | None = None,
    **kwargs,
) -> LazyKDE:
    """
    Read a Stereo-seq GEM file.

    Parameters
    ----------
    filepath : os.PathLike or str
        Path to the Stereo-seq file.
    resolution : float, optional
        Center-to-center distance of Stere-seq beads in nm, if None
        it will try to detect it from the chip definition in the file header
        if one exists.
    sep : str, optional
        Separator used in :py:func:`polars.read_csv`.
    n_threads : int, optional
        Number of threads used for reading and processing file. If `None` this will
        default to the number of available CPUs.
    kwargs
        Other keyword arguments will be passed to :py:func:`polars.read_csv`.

    Returns
    -------
    sainsc.LazyKDE
    """
    if n_threads is None:
        n_threads = _get_n_cpus()

    if resolution is None:
        chip = _get_stereo_header(filepath).get("StereoChip")
        if chip is not None:
            resolution = _get_stereo_resolution(chip)

    df = read_gem_file(filepath, sep=sep, n_threads=n_threads, **kwargs)

    counts = GridCounts.from_dataframe(
        df, binsize=None, resolution=resolution, n_threads=n_threads
    )

    return LazyKDE(counts, n_threads=n_threads)


def read_StereoSeq_bins(
    filepath: _PathLike,
    bin_size: int = 50,
    *,
    spatialdata: bool = False,
    resolution: float | None = None,
    sep: str = "\t",
    n_threads: int | None = None,
    **kwargs,
) -> "AnnData | SpatialData":
    """
    Read a Stereo-seq GEM file into bins.

    Parameters
    ----------
    filepath : os.PathLike or str
        Path to the Stereo-seq file.
    bin_size : int, optional
        Defines the size of bins along both dimensions
        e.g 50 will results in bins of size 50x50.
    spatialdata : bool, optional
        If True will load the data as a SpatialData object else as an AnnData object.
    resolution : float, optional
        Center-to-center distance of Stere-seq beads in nm, if None
        it will try to detect it from the chip definition in the file header
        if one exists.
    sep : str, optional
        Separator used in :py:func:`polars.read_csv`.
    n_threads : int, optional
        Number of threads used for reading and processing file. If `None` this will
        default to the number of available CPUs.
    kwargs
        Other keyword arguments will be passed to :py:func:`polars.read_csv`.

    Returns
    -------
    anndata.AnnData | spatialdata.SpatialData
        AnnData or SpatialData object of the bins with coordinates stored in
        :py:attr:`anndata.AnnData.obsm` with the key `'spatial'`.

    Raises
    ------
    ModuleNotFoundError
        If `spatialdata` is set to `True` but the package is not installed.
    """
    if n_threads is None:
        n_threads = _get_n_cpus()

    header = _get_stereo_header(filepath)
    df = read_gem_file(filepath, sep=sep, n_threads=n_threads, **kwargs)
    df = df.with_columns(pl.col(i) - pl.col(i).min() for i in ["x", "y"])
    df = _bin_coordinates(df.to_pandas(), bin_size)

    coord_codes, coordinates = _categorical_coordinate(
        df.pop("x").to_numpy(), df.pop("y").to_numpy(), n_threads=n_threads
    )

    # Duplicate entries in csr_matrix are summed which automatically gives bin merging
    counts = csr_matrix(
        (df.pop("count"), (coord_codes, df["gene"].cat.codes)),
        shape=(coordinates.shape[0], df["gene"].cat.categories.size),
        dtype=np.int32,
    )

    del coord_codes

    obs = pd.DataFrame(
        index=_get_coordinate_index(
            coordinates[:, 0], coordinates[:, 1], name="bin", n_threads=n_threads
        )
    )
    genes = pd.DataFrame(index=pd.Index(df["gene"].cat.categories, name="gene"))

    if resolution is None:
        chip = header.get("StereoChip")
        if chip is not None:
            resolution = _get_stereo_resolution(chip)

    adata = AnnData(
        X=counts,
        obs=obs,
        var=genes,
        obsm={"spatial": coordinates},
        uns={"file_header": header, "resolution": resolution, "bin_size": bin_size},
    )

    if spatialdata:
        try:
            from geopandas import GeoDataFrame
            from shapely import Polygon
            from spatialdata import SpatialData
            from spatialdata.models import ShapesModel, TableModel

            bin_name = f"bins{bin_size}"

            x, y = adata.obsm["spatial"].T
            del adata.obsm["spatial"]

            df = pd.DataFrame({"x": x, "y": y}, index=adata.obs_names).assign(
                x1=lambda df: df["x"] * bin_size,
                x2=lambda df: df["x1"] + bin_size,
                y1=lambda df: df["y"] * bin_size,
                y2=lambda df: df["y1"] + bin_size,
            )

            shapes = ShapesModel.parse(
                GeoDataFrame(
                    {
                        "geometry": df.apply(
                            lambda r: Polygon(
                                [
                                    (r["x1"], r["y1"]),
                                    (r["x1"], r["y2"]),
                                    (r["x2"], r["y2"]),
                                    (r["x2"], r["y1"]),
                                ]
                            ),
                            axis=1,
                        )
                    }
                )
            )

            adata.obs["region"] = bin_name
            adata.obs["region"] = adata.obs["region"].astype("category")
            adata.obs["instance_key"] = adata.obs_names
            table = TableModel.parse(
                adata, region=bin_name, region_key="region", instance_key="instance_key"
            )

            return SpatialData(
                shapes={bin_name: shapes}, tables={f"{bin_name}_annotation": table}
            )

        except ModuleNotFoundError as e:
            _raise_module_load_error(
                e, "read_StereoSeq_bins", pkg="spatialdata", extra="spatialdata"
            )

    else:
        return adata



from collections.abc import Callable

import anndata as ad
import pandas as pd


def celltype_signatures(
    adata: ad.AnnData,
    *,
    celltype_col: str = "leiden",
    agg_method: str | Callable = "mean",
) -> pd.DataFrame:
    """
    Calculate gene expression signatures per 'celltype'.

    Note, that this will make a dense copy of `adata.X` therefore potentially leading
    to large memory usage.

    Parameters
    ----------
    adata : anndata.AnnData
    celltype_col : str, optional
        Name of column in :py:attr:`anndata.AnnData.obs` containing cell-type
        information.
    agg_method : str or collections.abc.Callable, optional
        Function to aggregate gene expression per cluster used by
        :py:meth:`pandas.DataFrame.agg`.

    Returns
    -------
    pandas.DataFrame
        :py:class:`pandas.DataFrame` of gene expression aggregated per 'celltype'.
    """
    signatures = (
        adata.to_df()
        .merge(adata.obs[celltype_col], left_index=True, right_index=True)
        .groupby(celltype_col, observed=True, sort=False)
        .agg(agg_method)
        .transpose()
        .rename_axis(adata.var_names.name)
    )

    signatures /= signatures.sum(axis=0)

    return signatures



from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("sainsc")
except PackageNotFoundError:
    __version__ = "unknown version"

from ._utils_rust import GridCounts
from .io import read_StereoSeq, read_StereoSeq_bins
from .lazykde import LazyKDE, gaussian_kernel

__all__ = [
    "GridCounts",
    "LazyKDE",
    "gaussian_kernel",
    "read_StereoSeq",
    "read_StereoSeq_bins",
]



import gzip
from pathlib import Path
from typing import Literal

import numpy as np
from numpy.typing import NDArray

from .._typealias import _PathLike
from .._utils_rust import categorical_coordinate


def _categorical_coordinate(
    x: NDArray[np.int32], y: NDArray[np.int32], *, n_threads: int = 1
) -> tuple[NDArray[np.int32], NDArray[np.int32]]:
    assert len(x) == len(y)

    return categorical_coordinate(x, y, n_threads=n_threads)


# currently no need to support all file modes
_File_Mode = Literal["r", "w", "rb", "wb"]


def _open_file(file: _PathLike, mode: _File_Mode = "r"):
    file = Path(file)
    if file.suffix == ".gz":
        return gzip.open(file, mode)
    else:
        return open(file, mode)



from collections.abc import Iterable
from itertools import chain
from typing import TYPE_CHECKING, Any

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import polars as pl
import seaborn as sns
from anndata import AnnData
from matplotlib.axes import Axes
from matplotlib.colors import to_rgb
from matplotlib.figure import Figure
from matplotlib.patches import Patch
from matplotlib_scalebar.scalebar import ScaleBar
from mpl_toolkits import axes_grid1
from numba import njit
from numpy.typing import NDArray
from scipy.sparse import coo_array, csc_array, csr_array
from skimage.feature import peak_local_max

from .._typealias import _Cmap, _Csx, _CsxArray, _Local_Max, _RangeTuple2D
from .._utils import _get_n_cpus, _raise_module_load_error
from .._utils_rust import (
    GridCounts,
    cosinef32_and_celltypei8,
    cosinef32_and_celltypei16,
    kde_at_coord,
    sparse_kde_csx_py,
)
from ._kernel import gaussian_kernel
from ._utils import (
    _SCALEBAR,
    CosineCelltypeCallable,
    _apply_color,
    _filter_blobs,
    _get_cell_dtype,
    _localmax_anndata,
)

# from typing import Self

if TYPE_CHECKING:
    from spatialdata import SpatialData


class LazyKDE:
    """
    Class to analyze kernel density estimates (KDE) for large number of genes.

    The KDE of the genes will be calculated when needed to avoid storing large volumes
    of data in memory.
    """

    def __init__(
        self,
        counts: GridCounts,
        *,
        n_threads: int | None = None,
    ):
        """
        Parameters
        ----------
        counts : sainsc.GridCounts
            Gene counts.
        n_threads : int, optional
            Number of threads used for reading and processing file. If `None` this will
            default to the number of available CPUs.
        """
        if n_threads is None:
            n_threads = _get_n_cpus()

        self.counts: GridCounts = counts
        """
        sainsc.GridCounts :  Spatial gene counts.
        """

        self.counts.n_threads = n_threads

        self._threads = n_threads

        self._kernel: NDArray[np.float32] | None = None
        self._total_mRNA: NDArray[np.unsignedinteger] | None = None
        self._total_mRNA_KDE: NDArray[np.float32] | None = None
        self._background: NDArray[np.bool_] | None = None
        self._local_maxima: _Local_Max | None = None
        self._celltype_map: NDArray[np.signedinteger] | None = None
        self._cosine_similarity: NDArray[np.float32] | None = None
        self._assignment_score: NDArray[np.float32] | None = None
        self._celltypes: list[str] | None = None

    @classmethod
    def from_dataframe(
        cls, df: pl.DataFrame | pd.DataFrame, *, n_threads: int | None = None, **kwargs
    ):
        """
        Construct a LazyKDE from a DataFrame.

        The DataFrame must provide a 'gene', 'x', and 'y' column. If a 'count' column
        exists it will be used as counts else a count of 1 (single molecule) per row
        will be assumed.

        Parameters
        ----------
        df : polars.DataFrame | pandas.DataFrame
        n_threads : int, optional
            Number of threads used for reading and processing file. If `None` this will
            default to the number of available CPUs.
        kwargs
            Other keyword arguments are passed to
            :py:meth:`sainsc.GridCounts.from_dataframe`.
        """

        count_col = ["count"] if "count" in df.columns else []
        columns = ["gene", "x", "y"] + count_col

        if isinstance(df, pd.DataFrame):
            df = pl.from_pandas(df[columns])
        else:
            df = df.select(pl.col(columns))

        return cls(
            GridCounts.from_dataframe(df, n_threads=n_threads, **kwargs),
            n_threads=n_threads,
        )

    ## Kernel
    def gaussian_kernel(
        self,
        bw: float,
        *,
        unit: str = "px",
        truncate: float = 2,
        circular: bool = False,
    ):
        """
        Set the kernel used for kernel density estimation (KDE) to gaussian.

        Parameters
        ----------
        bw : float
            Bandwidth of the kernel.
        unit : str
            Which unit the bandwidth of the kernel is defined in: 'px' or 'um'.
            'um' requires :py:attr:`sainsc.LazyKDE.resolution` to be set correctly.
        truncate : float, optional
            The radius for calculating the KDE is calculated as `bw` * `truncate`.
            Refer to :py:func:`scipy.ndimage.gaussian_filter`.
        circular : bool, optional
            If `True` calculate the KDE using a circular kernel instead of square by
            setting all values outside the radius `bw` * `truncate` to 0.

        Raises
        ------
        ValueError
            If `unit` is neither 'px' nor 'um'.
        ValueError
            If `unit` is 'um' but `resolution` is not set.

        See Also
        --------
        :py:meth:`sainsc.LazyKDE.kde`
        """

        if unit == "um":
            if self.resolution is None:
                raise ValueError(
                    "Using `unit`='um' requires the `resolution` to be set."
                )
            bw /= self.resolution / 1_000
        elif unit != "px":
            raise ValueError("`unit` must be either 'px' or 'um'")
        dtype = np.float32
        radius = round(truncate * bw)
        self.kernel = gaussian_kernel(bw, radius, dtype=dtype, circular=circular)

    ## KDE
    def kde(self, gene: str, *, threshold: float | None = None) -> _CsxArray:
        """
        Calculate kernel density estimate (KDE).

        The kernel will be used from :py:attr:`sainsc.LazyKDE.kernel`.

        Parameters
        ----------
        gene : collections.abc.Sequence[str]
            List of genes for which to calculate the KDE.
        threshold : float, optional
            Relative threshold of maximum of the kernel that is used to filter beads.
            All values below :math:`threshold * max(kernel)` are set to 0. Filtering is done
            after calculating the KDE which sets it apart from reducing `truncate`.

        See Also
        --------
        :py:meth:`sainsc.LazyKDE.gaussian_kernel`
        """
        return self._kde(self.counts[gene], threshold)

    def _kde(self, arr: NDArray | _Csx, threshold: float | None = None) -> _CsxArray:
        if self.kernel is None:
            raise ValueError("`kernel` must be set before running KDE")

        if threshold is None:
            threshold = 0

        if isinstance(arr, np.ndarray):
            # scipy.ndimage.convolve could be used for dense arrays but seems to be
            # slower than converting to sparse and running custom kde
            arr = csr_array(arr)

        if arr.dtype == np.uint32:
            return sparse_kde_csx_py(arr, self.kernel, threshold=threshold)
        else:
            raise TypeError("Sparse KDE currently only supports 'numpy.uint32'")

    def calculate_total_mRNA(self):
        """
        Calculate kernel density estimate (KDE) for the total mRNA.

        See Also
        --------
        :py:meth:`sainsc.LazyKDE.calculate_total_mRNA_KDE`
        """

        self._total_mRNA = self.counts.grid_counts()

    def calculate_total_mRNA_KDE(self):
        """
        Calculate kernel density estimate (KDE) for the total mRNA.

        If :py:attr:`sainsc.LazyKDE.total_mRNA` has not been calculated
        :py:meth:`sainsc.LazyKDE.calculate_total_mRNA` is run first.

        Raises
        ------
        ValueError
            If `self.kernel` is not set.

        See Also
        --------
        :py:meth:`sainsc.LazyKDE.gaussian_kernel`
        :py:meth:`sainsc.LazyKDE.kde`
        """
        if self.total_mRNA is None or self.total_mRNA.shape != self.shape:
            self.calculate_total_mRNA()
        assert self.total_mRNA is not None
        self._total_mRNA_KDE = self._kde(self.total_mRNA).toarray()

    ## Local maxima / cell proxies
    def find_local_maxima(self, min_dist: int, min_area: int = 0):
        """
        Find the local maxima of the kernel density estimates.

        The local maxima are detected from the KDE of the total mRNA stored in
        :py:attr:`sainsc.LazyKDE.total_mRNA_KDE`. Background as defined in
        :py:attr:`sainsc.LazyKDE.background` will be removed before identifying
        local maxima.

        Parameters
        ----------
        min_dist : int
            Minimum distance between two maxima in pixels.
        min_area : int, optional
            Minimum area of connected pixels that are not background to be
            considered for maxima detection. Allows ignoring maxima in noisy spots.
        """
        if self.total_mRNA_KDE is None:
            raise ValueError(
                "`total_mRNA_KDE` must be calculated before finding local maxima"
            )

        if self.background is not None:
            foreground = ~self.background
            if min_area > 0:
                foreground = _filter_blobs(foreground, min_area)
        else:
            foreground = None

        local_max = peak_local_max(
            self.total_mRNA_KDE,
            min_distance=min_dist,
            exclude_border=False,
            labels=foreground,
        )

        self._local_maxima = (local_max[:, 0], local_max[:, 1])

    def load_local_maxima(
        self, genes: Iterable[str] | None = None, *, spatialdata: bool = False
    ) -> "AnnData | SpatialData":
        """
        Load the gene expression (KDE) of the local maxima.

        The local maxima (:py:attr:`sainsc.LazyKDE.local_maxima`) are calculated and
        returned as :py:class:`anndata.AnnData` object.

        Parameters
        ----------
        genes : collections.abc.Iterable[str], optional
            List of genes for which the KDE will be calculated.
        spatialdata : bool, optional
            If True will load the data as a SpatialData object including the totalRNA
            projection and cell-type map if available. If False an AnnData object is
            returned.

        Returns
        -------
        anndata.AnnData | spatialdata.SpatialData

        Raises
        ------
        ModuleNotFoundError
            If `spatialdata` is set to `True` but the package is not installed.
        ValueError
            If `self.kernel` is not set.
        """
        if self.local_maxima is None:
            raise ValueError("`local_maxima` have to be identified before loading")

        if genes is None:
            genes = self.genes
        else:
            genes = list(genes)

        kde = self._load_KDE_maxima(genes)
        adata = _localmax_anndata(
            kde, genes, self.local_maxima, name="local_maxima", n_threads=self.n_threads
        )

        if spatialdata:
            try:
                from spatialdata import SpatialData
                from spatialdata.models import (
                    Image2DModel,
                    Labels2DModel,
                    PointsModel,
                    TableModel,
                )

                x, y = adata.obsm["spatial"].T
                del adata.obsm["spatial"]

                localmax_name = "local_maxima"

                local_max = PointsModel.parse(
                    pd.DataFrame({"x": x, "y": y}, index=adata.obs_names)
                )

                adata.obs["region"] = localmax_name
                adata.obs["region"] = adata.obs["region"].astype("category")
                adata.obs["instance_key"] = adata.obs_names

                local_max_anno = TableModel.parse(
                    adata,
                    region=localmax_name,
                    region_key="region",
                    instance_key="instance_key",
                )

                sdata_dict: dict[str, Any] = {
                    localmax_name: local_max,
                    f"{localmax_name}_annotation": local_max_anno,
                }

                if self.total_mRNA_KDE is not None:

                    sdata_dict["total_mRNA"] = Image2DModel.parse(
                        np.atleast_3d(self.total_mRNA_KDE).T, dims=("c", "y", "x")
                    )

                if self.celltype_map is not None:
                    label_name = "celltype_map"

                    labels = self.celltype_map + 1
                    if self.background is not None:
                        labels[self.background] = 0

                    sdata_dict[label_name] = Labels2DModel.parse(
                        labels.T, dims=("y", "x")
                    )

                    obs = pd.DataFrame(
                        {"region": label_name, "instance_key": self.celltypes},
                        index=self.celltypes,
                    ).astype({"region": "category"})

                    sdata_dict[f"{label_name}_annotation"] = TableModel.parse(
                        AnnData(obs=obs),
                        region=label_name,
                        region_key="region",
                        instance_key="instance_key",
                    )

                return SpatialData.from_elements_dict(sdata_dict)

            except ModuleNotFoundError as e:
                _raise_module_load_error(
                    e, "load_local_maxima", pkg="spatialdata", extra="spatialdata"
                )

        else:
            load_attr = [
                "total_mRNA_KDE",
                "cosine_similarity",
                "assignment_score",
                "celltype_map",
            ]
            x, y = adata.obsm["spatial"].T
            for name in load_attr:
                if (attr := getattr(self, name)) is not None:
                    # assert isinstance(attr, np.ndarray)
                    values = attr[x, y]
                    if name == "celltype_map":
                        assert self.celltypes is not None
                        adata.obs["celltype"] = pd.Categorical.from_codes(
                            values, self.celltypes
                        )
                    else:
                        adata.obs[name] = values

            return adata

    def _load_KDE_maxima(self, genes: list[str]) -> csc_array | csr_array:

        assert self.local_maxima is not None
        if self.kernel is None:
            raise ValueError("`kernel` must be set before running KDE")

        return kde_at_coord(
            self.counts, genes, self.kernel, self.local_maxima, n_threads=self.n_threads
        )

    ## Celltyping
    def filter_background(
        self,
        min_norm: float | dict[str, float],
        min_cosine: float | dict[str, float] | None = None,
        min_assignment: float | dict[str, float] | None = None,
    ):
        """
        Define pixels as background.

        If using multiple thresholds (e.g. on norm and cosine similarity) they will be
        combined and pixels are defined as background if they are lower than any of the
        thresholds.

        Parameters
        ----------
        min_norm : float or dict[str, float]
            The threshold for defining background based on
            :py:attr:`sainsc.LazyKDE.total_mRNA_KDE`.
            Either a float which is used as global threshold or a mapping from cell types
            to thresholds. Cell-type assignment is needed for cell type-specific thresholds.
        min_cosine : float or dict[str, float], optional
            The threshold for defining background based on
            :py:attr:`sainsc.LazyKDE.cosine_similarity`. Cell type-specific thresholds
            can be defined as for `min_norm`.
        min_assignment : float or dict[str, float], optional
            The threshold for defining background based on
            :py:attr:`sainsc.LazyKDE.assignment_score`. Cell type-specific thresholds
            can be defined as for `min_norm`.

        Raises
        ------
        ValueError
            If cell type-specific thresholds do not include all cell types or if
            using cell type-specific thresholds before cell type assignment.
        """

        @njit
        def _map_celltype_to_value(
            ct_map: NDArray[np.integer], thresholds: tuple[float, ...]
        ) -> NDArray[np.floating]:
            values = np.zeros(shape=ct_map.shape, dtype=float)
            for i in range(ct_map.shape[0]):
                for j in range(ct_map.shape[1]):
                    if ct_map[i, j] >= 0:
                        values[i, j] = thresholds[ct_map[i, j]]
            return values

        if self.total_mRNA_KDE is None:
            raise ValueError(
                "`total_mRNA_KDE` needs to be calculated before filtering background"
            )

        if isinstance(min_norm, dict):
            if self.celltypes is None or self.celltype_map is None:
                raise ValueError(
                    "Cell type-specific threshold can only be used after cell-type assignment"
                )
            elif not all([ct in min_norm.keys() for ct in self.celltypes]):
                raise ValueError("'min_norm' does not contain all celltypes.")
            idx2threshold = tuple(min_norm[ct] for ct in self.celltypes)
            threshold = _map_celltype_to_value(self.celltype_map, idx2threshold)
            background = self.total_mRNA_KDE < threshold
        else:
            background = self.total_mRNA_KDE < min_norm

        if min_cosine is not None:
            if self.cosine_similarity is None:
                raise ValueError(
                    "Cosine similarity threshold can only be used after cell-type assignment"
                )
            if isinstance(min_cosine, dict):
                if self.celltypes is None or self.celltype_map is None:
                    raise ValueError(
                        "Cell type-specific threshold can only be used after cell-type assignment"
                    )
                elif not all([ct in min_cosine.keys() for ct in self.celltypes]):
                    raise ValueError("'min_cosine' does not contain all celltypes.")
                idx2threshold = tuple(min_cosine[ct] for ct in self.celltypes)
                threshold = _map_celltype_to_value(self.celltype_map, idx2threshold)
                background |= self.cosine_similarity <= threshold
            else:
                background |= self.cosine_similarity <= min_cosine

        if min_assignment is not None:
            if self.assignment_score is None:
                raise ValueError(
                    "Assignment score threshold can only be used after cell-type assignment"
                )
            if isinstance(min_assignment, dict):
                if self.celltypes is None or self.celltype_map is None:
                    raise ValueError(
                        "Cell type-specific threshold can only be used after cell-type assignment"
                    )
                elif not all([ct in min_assignment.keys() for ct in self.celltypes]):
                    raise ValueError("'min_assignment' does not contain all celltypes.")
                idx2threshold = tuple(min_assignment[ct] for ct in self.celltypes)
                threshold = _map_celltype_to_value(self.celltype_map, idx2threshold)
                background |= self.assignment_score <= threshold
            else:
                background |= self.assignment_score <= min_assignment

        self._background = background

    @staticmethod
    def _calculate_cosine_celltype_fn(dtype) -> CosineCelltypeCallable:
        if dtype == np.int8:
            return cosinef32_and_celltypei8
        elif dtype == np.int16:
            return cosinef32_and_celltypei16
        else:
            raise NotImplementedError

    def assign_celltype(
        self,
        signatures: pd.DataFrame,
        *,
        log: bool = False,
        chunk: tuple[int, int] = (500, 500),
    ):
        """
        Calculate the cosine similarity with known cell-type signatures.

        For each bead calculate the cosine similarity with a set of cell-type signatures.
        The cell-type with highest score will be assigned to the corresponding bead.

        Parameters
        ----------
        signatures : pandas.DataFrame
            DataFrame of cell-type signatures. Columns are cell-types and index are genes.
        log : bool
            Whether to log transform the KDE when calculating the cosine similarity.
            This is useful if the gene signatures are derived from log-transformed data.
        chunk : tuple[int, int]
            Size of the chunks for processing. Larger chunks require more memory but
            have less duplicated computation.

        Raises
        ------
        ValueError
            If not all genes of the `signatures` are available.
        ValueError
            If `self.kernel` is not set.
        ValueError
            If `chunk` is smaller than the shape of `self.kernel`.
        """

        if not all(signatures.index.isin(self.genes)):
            raise ValueError(
                "Not all genes in the gene signature are part of this KDE."
            )

        if self.kernel is None:
            raise ValueError("`kernel` must be set before running KDE")

        if not all(s < c for s, c in zip(self.kernel.shape, chunk)):
            raise ValueError("`chunk` must be larger than shape of kernel.")

        dtype = np.float32

        celltypes = signatures.columns.tolist()
        ct_dtype = _get_cell_dtype(len(celltypes))

        # scale signatures to unit norm
        signatures_mat = signatures.to_numpy()
        signatures_mat = (
            signatures_mat / np.linalg.norm(signatures_mat, axis=0)
        ).astype(dtype, copy=False)

        genes = signatures.index.to_list()

        fn = self._calculate_cosine_celltype_fn(ct_dtype)

        self._cosine_similarity, self._assignment_score, self._celltype_map = fn(
            self.counts,
            genes,
            signatures_mat,
            self.kernel,
            log=log,
            chunk_size=chunk,
            n_threads=self.n_threads,
        )
        self._celltypes = celltypes

    ## Plotting
    def _plot_2d(
        self,
        img: NDArray,
        title: str,
        *,
        remove_background: bool = False,
        crop: _RangeTuple2D | None = None,
        scalebar: bool = True,
        im_kwargs: dict = dict(),
        scalebar_kwargs: dict = _SCALEBAR,
    ) -> Figure:
        if remove_background:
            if self.background is not None:
                img = img.copy()
                img[self.background] = 0
            else:
                raise ValueError("`background` is undefined")

        if crop is not None:
            img = img[tuple(slice(*c) for c in crop)]
        fig, ax = plt.subplots(1, 1)
        assert isinstance(ax, Axes)
        im = ax.imshow(img.T, origin="lower", **im_kwargs)
        ax.set_title(title)

        divider = axes_grid1.make_axes_locatable(ax)
        cax = divider.append_axes("right", size="3%", pad=0.1)
        fig.colorbar(im, cax=cax)

        if scalebar:
            self._add_scalebar(ax, **scalebar_kwargs)
        return fig

    def _add_scalebar(self, ax: Axes, **kwargs):
        if self.counts.resolution is None:
            raise ValueError("'resolution' must be set when using scalebar")
        ax.add_artist(ScaleBar(self.counts.resolution, units="nm", **kwargs))

    def plot_genecount_histogram(self, **kwargs) -> Figure:
        """
        Plot a histogram of the counts per gene.

        Parameters
        ----------
        kwargs
            Other keyword arguments are passed to :py:func:`seaborn.histplot`

        Returns
        -------
        matplotlib.figure.Figure
        """
        fig, ax = plt.subplots(1, 1)
        assert isinstance(ax, Axes)
        sns.histplot(
            np.fromiter(self.counts.gene_counts().values(), dtype=int),
            log_scale=True,
            ax=ax,
            **kwargs,
        )
        ax.set(xlabel="Counts per gene", ylabel="# genes")
        return fig

    def plot_KDE_histogram(
        self,
        *,
        gene: str | None = None,
        remove_background: bool = False,
        crop: _RangeTuple2D | None = None,
        **kwargs,
    ) -> Figure:
        """
        Plot a histogram of the kernel density estimates.

        Plots either the kernel density estimate (KDE) of the total mRNA
        (:py:attr:`sainsc.LazyKDE.total_mRNA_KDE`) or of a single gene if `gene` is
        provided.

        Parameters
        ----------
        gene : str, optional
            Gene for which the KDE histogram is plotted.
        remove_background : bool, optional
            If `True`, all pixels for which :py:attr:`sainsc.LazyKDE.background` is
            `False` are set to 0.
        crop : tuple[tuple[int, int], tuple[int, int]], optional
            Coordinates to crop the data defined as `((xmin, xmax), (ymin, ymax))`.
        kwargs
            Other keyword arguments are passed to :py:func:`matplotlib.pyplot.hist`.

        Returns
        -------
        matplotlib.figure.Figure

        See Also
        --------
        :py:meth:`sainsc.LazyKDE.kde`
        """
        name = "total mRNA" if gene is None else gene

        if gene is not None:
            kde = self.kde(gene)
        else:
            if self.total_mRNA_KDE is not None:
                kde = self.total_mRNA_KDE
            else:
                raise ValueError("`total_mRNA_KDE` has not been calculated")

        if remove_background:
            if self.background is not None:
                kde[self.background] = 0
            else:
                raise ValueError("`background` is undefined")

        if crop is not None:
            kde = kde[tuple(slice(*c) for c in crop)]

        fig, ax = plt.subplots(1, 1)
        assert isinstance(ax, Axes)
        ax.hist(coo_array(kde).data, **kwargs)
        ax.set(xlabel=f"KDE of {name}", ylabel="# pixels")
        return fig

    def plot_genecount(
        self,
        *,
        gene: str | None = None,
        crop: _RangeTuple2D | None = None,
        scalebar: bool = True,
        im_kwargs: dict = dict(),
        scalebar_kwargs: dict = _SCALEBAR,
    ) -> Figure:
        """
        Plot the gene expression counts.

        By default this will plot the :py:attr:`sainsc.LazyKDE.total_mRNA`. If
        `gene` is specified the respective gene will be plotted.

        Parameters
        ----------
        gene : str, optional
            Gene in :py:attr:`sainsc.LazyKDE.genes` to use for plotting.
        crop : tuple[tuple[int, int], tuple[int, int]], optional
            Coordinates to crop the data defined as `((xmin, xmax), (ymin, ymax))`.
        scalebar : bool, optional
            If `True`, add a ``matplotlib_scalebar.scalebar.ScaleBar`` to the plot.
        im_kwargs : dict[str, typing.Any], optional
            Keyword arguments that are passed to :py:func:`matplotlib.pyplot.imshow`.
        scalebar_kwargs : dict[str, typing.Any], optional
            Keyword arguments that are passed to ``matplotlib_scalebar.scalebar.ScaleBar``.

        Returns
        -------
        matplotlib.figure.Figure

        Raises
        ------
        ValueError
            If :py:attr:`sainsc.LazyKDE.total_mRNA` has not been calculated.
        """
        if gene is not None:
            img = self.counts[gene].toarray()
        else:
            if self.total_mRNA is not None:
                img = self.total_mRNA
            else:
                raise ValueError(
                    "`total_mRNA` has not been calculated."
                    "Run `calculate_total_mRNA` first."
                )
        title = "total mRNA" if gene is None else gene

        return self._plot_2d(
            img,
            title,
            crop=crop,
            scalebar=scalebar,
            im_kwargs=im_kwargs,
            scalebar_kwargs=scalebar_kwargs,
        )

    def plot_KDE(
        self,
        *,
        gene: str | None = None,
        remove_background: bool = False,
        crop: _RangeTuple2D | None = None,
        scalebar: bool = True,
        im_kwargs: dict = dict(),
        scalebar_kwargs: dict = _SCALEBAR,
    ) -> Figure:
        """
        Plot the kernel density estimate (KDE).

        By default this will plot the KDE of the total mRNA
        (:py:attr:`sainsc.LazyKDE.total_mRNA_KDE`). If `gene` is specified the
        respective KDE will be computed and plotted.

        Parameters
        ----------
        gene : str, optional
            Gene in :py:attr:`sainsc.LazyKDE.genes` to use for plotting.
        remove_background : bool, optional
            If `True`, all pixels for which :py:attr:`sainsc.LazyKDE.background` is
            `False` are set to 0.
        crop : tuple[tuple[int, int], tuple[int, int]], optional
            Coordinates to crop the data defined as `((xmin, xmax), (ymin, ymax))`.
        scalebar : bool, optional
            If `True`, add a ``matplotlib_scalebar.scalebar.ScaleBar`` to the plot.
        im_kwargs : dict[str, typing.Any], optional
            Keyword arguments that are passed to :py:func:`matplotlib.pyplot.imshow`.
        scalebar_kwargs : dict[str, typing.Any], optional
            Keyword arguments that are passed to ``matplotlib_scalebar.scalebar.ScaleBar``.

        Returns
        -------
        matplotlib.figure.Figure

        Raises
        ------
        ValueError
            If :py:attr:`sainsc.LazyKDE.total_mRNA_KDE` has not been calculated.

        See Also
        --------
        :py:meth:`sainsc.LazyKDE.kde`
        """
        if gene is not None:
            img = self.kde(gene).toarray()
        else:
            if self.total_mRNA_KDE is not None:
                img = self.total_mRNA_KDE
            else:
                raise ValueError(
                    "`total_mRNA_KDE` has not been calculated."
                    "Run `calculate_total_mRNA_KDE` first."
                )

        title = "KDE of " + ("total mRNA" if gene is None else gene)

        return self._plot_2d(
            img,
            title,
            remove_background=remove_background,
            crop=crop,
            scalebar=scalebar,
            im_kwargs=im_kwargs,
            scalebar_kwargs=scalebar_kwargs,
        )

    def plot_local_maxima(
        self,
        *,
        crop: _RangeTuple2D | None = None,
        background_kwargs: dict = dict(),
        scatter_kwargs: dict = dict(),
    ) -> Figure:
        """
        Plot the local kernel density estimate maxima.

        Parameters
        ----------
        crop : tuple[tuple[int, int], tuple[int, int]], optional
            Coordinates to crop the data defined as `((xmin, xmax), (ymin, ymax))`.
        background_kwargs : dict, optional
            Keyword arguments that are passed to :py:meth:`sainsc.LazyKDE.plot_KDE`.
        scatter_kwargs : dict, optional
            Keyword arguments that are passed to :py:func:`matplotlib.pyplot.scatter`.

        Returns
        -------
        matplotlib.figure.Figure

        See Also
        --------
        :py:meth:`sainsc.LazyKDE.find_local_maxima`
        """
        if self.local_maxima is None:
            raise ValueError

        x, y = self.local_maxima

        if crop is not None:
            x_min, x_max = crop[0]
            y_min, y_max = crop[1]
            keep = (x >= x_min) & (y >= y_min) & (x < x_max) & (y < y_max)
            x = x[keep] - x_min
            y = y[keep] - y_min

        fig = self.plot_KDE(crop=crop, **background_kwargs)
        fig.axes[0].scatter(x, y, **scatter_kwargs)
        return fig

    def plot_celltype_map(
        self,
        *,
        remove_background: bool = True,
        crop: _RangeTuple2D | None = None,
        scalebar: bool = True,
        cmap: _Cmap = "hls",
        background: str | tuple = "black",
        undefined: str | tuple = "grey",
        scalebar_kwargs: dict = _SCALEBAR,
        return_img: bool = False,
    ) -> Figure | NDArray[np.uint8]:
        """
        Plot the cell-type annotation.

        Parameters
        ----------
        remove_background : bool, optional
            If `True`, all pixels for which :py:attr:`sainsc.LazyKDE.background` is
            `False` are set to 0.
        crop : tuple[tuple[int, int], tuple[int, int]], optional
            Coordinates to crop the data defined as `((xmin, xmax), (ymin, ymax))`.
        scalebar : bool, optional
            If `True`, add a ``matplotlib_scalebar.scalebar.ScaleBar`` to the plot.
        cmap : str or list or dict, optional
            If it is a string it must be the name of a `cmap` that can be used in
            :py:func:`seaborn.color_palette`.
            If it is a list of colors it must have the same length as the number of
            celltypes.
            If it is a dictionary it must be a mapping from celltpye to color. Undefined
            celltypes are plotted according to `undefined`.
            Colors can either be provided as string that can be converted via
            :py:func:`matplotlib.colors.to_rgb` or as ``(r, g, b)``-tuple between 0-1.
        background : str | tuple[float, float, float]
            Color for the background.
        undefined : str | tuple[float, float, float]
            Color used for celltypes without a defined color.
        scalebar_kwargs : dict[str, typing.Any], optional
            Keyword arguments that are passed to ``matplotlib_scalebar.scalebar.ScaleBar``.
        return_img : bool, optional
            Return the cell-type map as 3D-array (x, y, RGB) instead of the Figure.

        Returns
        -------
        matplotlib.figure.Figure | numpy.ndarray[numpy.ubyte]


        See Also
        --------
        :py:meth:`sainsc.LazyKDE.assign_celltype`
        """
        if self.celltypes is None or self.celltype_map is None:
            raise ValueError("celltype assignment missing")

        n_celltypes = len(self.celltypes)

        celltype_map = self.celltype_map.copy()
        if remove_background:
            if self.background is None:
                raise ValueError("Background has not been filtered.")
            else:
                celltype_map[self.background] = -1

        if crop is not None:
            celltype_map = celltype_map[tuple(slice(*c) for c in crop)]

        # shift so 0 will be background
        celltype_map += 1

        if isinstance(cmap, str):
            color_map = sns.color_palette(cmap, n_colors=n_celltypes)
            assert isinstance(color_map, Iterable)
        else:
            if isinstance(cmap, list):
                if len(cmap) != n_celltypes:
                    raise ValueError("You need to provide 1 color per celltype")

            elif isinstance(cmap, dict):
                cmap = [cmap.get(cell, undefined) for cell in self.celltypes]

            color_map = [to_rgb(c) if isinstance(c, str) else c for c in cmap]

        # convert to uint8 to reduce memory of final image
        color_map_int = tuple(
            (np.array(c) * 255).round().astype(np.uint8)
            for c in chain([to_rgb(background)], color_map)
        )
        img = _apply_color(celltype_map.T, color_map_int)

        if return_img:
            return img

        legend_elements = [
            Patch(color=c, label=lbl) for c, lbl in zip(color_map, self.celltypes)
        ]

        fig, ax = plt.subplots()
        assert isinstance(ax, Axes)
        ax.imshow(img, origin="lower")
        ax.legend(
            title="Cell type",
            handles=legend_elements,
            ncols=-(n_celltypes // -20),  # ceildiv
            loc="center left",
            bbox_to_anchor=(1, 0.5),
        )

        if scalebar:
            self._add_scalebar(ax, **scalebar_kwargs)

        return fig

    def plot_cosine_similarity(
        self,
        *,
        remove_background: bool = False,
        crop: _RangeTuple2D | None = None,
        scalebar: bool = True,
        im_kwargs: dict = dict(),
        scalebar_kwargs: dict = _SCALEBAR,
    ) -> Figure:
        """
        Plot the cosine similarity from cell-type assignment.

        Parameters
        ----------
        remove_background : bool, optional
            If `True`, all pixels for which :py:attr:`sainsc.LazyKDE.background` is
            `False` are set to 0.
        crop : tuple[tuple[int, int], tuple[int, int]], optional
            Coordinates to crop the data defined as `((xmin, xmax), (ymin, ymax))`.
        scalebar : bool, optional
            If `True`, add a ``matplotlib_scalebar.scalebar.ScaleBar`` to the plot.
        im_kwargs : dict[str, typing.Any], optional
            Keyword arguments that are passed to :py:func:`matplotlib.pyplot.imshow`.
        scalebar_kwargs : dict[str, typing.Any], optional
            Keyword arguments that are passed to ``matplotlib_scalebar.scalebar.ScaleBar``.

        Returns
        -------
        matplotlib.figure.Figure

        See Also
        --------
        :py:meth:`sainsc.LazyKDE.assign_celltype`
        """
        if self.cosine_similarity is not None:
            img = self.cosine_similarity
        else:
            raise ValueError("Cell types have not been assigned")

        return self._plot_2d(
            img,
            "Cosine similarity",
            remove_background=remove_background,
            crop=crop,
            scalebar=scalebar,
            im_kwargs=im_kwargs,
            scalebar_kwargs=scalebar_kwargs,
        )

    def plot_assignment_score(
        self,
        *,
        remove_background: bool = False,
        crop: _RangeTuple2D | None = None,
        scalebar: bool = True,
        im_kwargs: dict = dict(),
        scalebar_kwargs: dict = _SCALEBAR,
    ) -> Figure:
        """
        Plot the assignment score from cell-type assignment.

        Parameters
        ----------
        remove_background : bool, optional
            If `True`, all pixels for which :py:attr:`sainsc.LazyKDE.background` is
            `False` are set to 0.
        crop : tuple[tuple[int, int], tuple[int, int]], optional
            Coordinates to crop the data defined as `((xmin, xmax), (ymin, ymax))`.
        scalebar : bool, optional
            If `True`, add a ``matplotlib_scalebar.scalebar.ScaleBar`` to the plot.
        im_kwargs : dict[str, typing.Any], optional
            Keyword arguments that are passed to :py:func:`matplotlib.pyplot.imshow`.
        scalebar_kwargs : dict[str, typing.Any], optional
            Keyword arguments that are passed to ``matplotlib_scalebar.scalebar.ScaleBar``.

        Returns
        -------
        matplotlib.figure.Figure

        See Also
        --------
        :py:meth:`sainsc.LazyKDE.assign_celltype`
        """
        if self.assignment_score is not None:
            img = self.assignment_score
        else:
            raise ValueError("Cell types have not been assigned")

        return self._plot_2d(
            img,
            "Assignment score",
            remove_background=remove_background,
            crop=crop,
            scalebar=scalebar,
            im_kwargs=im_kwargs,
            scalebar_kwargs=scalebar_kwargs,
        )

    ## Attributes
    @property
    def n_threads(self) -> int:
        """
        int: Number of threads that will be used for computations.

        Raises
        ------
            TypeError
                If setting with a type other than `int` or less than 0.
        """
        return self._threads

    @n_threads.setter
    def n_threads(self, n_threads: int):
        if isinstance(n_threads, int) and n_threads >= 0:
            if n_threads == 0:
                n_threads = _get_n_cpus()
            self._threads = n_threads
            self.counts.n_threads = n_threads
        else:
            raise TypeError("`n_threads` must be an `int` >= 0.")

    @property
    def shape(self) -> tuple[int, int]:
        """
        tuple[int, int]: Shape of the sample.
        """
        return self.counts.shape

    @property
    def genes(self) -> list[str]:
        """
        list[str]: List of genes.
        """
        return self.counts.genes()

    @property
    def resolution(self) -> float | None:
        """
        float: Resolution in nm / pixel.

        Raises
        ------
            TypeError
                If setting with a type other than `float` or `int`.
        """
        return self.counts.resolution

    @resolution.setter
    def resolution(self, resolution: float):
        self.counts.resolution = resolution

    @property
    def kernel(self) -> np.ndarray | None:
        """
        numpy.ndarray: Map of the KDE of total mRNA.

        Raises
        ------
            ValueError
                If kernel is not a square, 2D :py:class:`numpy.ndarray` of uneven length.
        """
        return self._kernel

    @kernel.setter
    def kernel(self, kernel: np.ndarray):
        if (
            len(kernel.shape) != 2
            or kernel.shape[0] != kernel.shape[1]
            or any(i % 2 == 0 for i in kernel.shape)
        ):
            raise ValueError(
                "`kernel` currently only supports 2D squared arrays of uneven length."
            )
        else:
            self._kernel = kernel.astype(np.float32)

    @property
    def local_maxima(self) -> _Local_Max | None:
        """
        tuple[numpy.ndarray[numpy.signedinteger], ...]: Coordinates of local maxima.
        """
        return self._local_maxima

    @property
    def total_mRNA(self) -> NDArray[np.unsignedinteger] | None:
        """
        numpy.ndarray[numpy.unsignedinteger]: Map of the total mRNA.
        """
        return self._total_mRNA

    @property
    def total_mRNA_KDE(self) -> NDArray[np.single] | None:
        """
        numpy.ndarray[numpy.single]: Map of the KDE of total mRNA.
        """
        return self._total_mRNA_KDE

    @property
    def background(self) -> NDArray[np.bool_] | None:
        """
        numpy.ndarray[numpy.bool]: Map of pixels that are assigned as background.

        Raises
        ------
            TypeError
                If setting with array that is not of type `numpy.bool`.
            ValueError
                If setting with array that has different shape than `self`.
        """
        return self._background

    @background.setter
    def background(self, background: NDArray[np.bool_]):
        if background.shape != self.shape:
            raise ValueError("`background` must have same shape as `self`")
        else:
            self._background = background

    @property
    def celltypes(self) -> list[str] | None:
        """
        list[str]: List of assigned celltypes.
        """
        return self._celltypes

    @property
    def cosine_similarity(self) -> NDArray[np.single] | None:
        """
        numpy.ndarray[numpy.single]: Cosine similarity for each pixel.
        """
        return self._cosine_similarity

    @property
    def assignment_score(self) -> NDArray[np.single] | None:
        """
        numpy.ndarray[numpy.single]: Assignment score for each pixel.

        Let `x` be the gene expression of a pixel, and `i` and `j` the signatures of the
        best and 2nd best scoring cell type, respectively. The assignment score is
        calculated as :math:`\\frac{cos(\\theta_{xi}) - cos(\\theta_{xj})}{cos(\\pi/2 - \\theta_{ij})}`
        where :math:`\\theta` is the angle between the corresponding vectors.
        """
        return self._assignment_score

    @property
    def celltype_map(self) -> NDArray[np.signedinteger] | None:
        """
        numpy.ndarray[numpy.signedinteger]: Cell-type map of cell-type indices.

        Each number corresponds to the index in :py:attr:`sainsc.LazyKDE.celltypes`,
        and -1 to unassigned (background).
        """
        return self._celltype_map

    def __str__(self) -> str:
        repr = [
            f"LazyKDE ({self.n_threads} threads)",
            f"genes: {len(self.genes)}",
            f"shape: {self.shape}",
        ]
        if self.resolution is not None:
            repr.append(f"resolution: {self.resolution} nm / px")
        if self.kernel is not None:
            repr.append(f"kernel: {self.kernel.shape}")
        if self.background is not None:
            repr.append("background: set")
        if self.local_maxima is not None:
            repr.append(f"local maxima: {len(self.local_maxima[0])}")
        if self.celltypes is not None:
            repr.append(f"celltypes: {len(self.celltypes)}")

        spacing = "    "

        return f"\n{spacing}".join(repr)



import math
from typing import TypeVar

import numpy as np
from numpy.typing import DTypeLike, NDArray
from scipy import ndimage, signal

T = TypeVar("T", bound=np.number)


def _make_circular_mask(radius: int) -> NDArray[np.bool_]:
    diameter = radius * 2 + 1
    x, y = np.ogrid[:diameter, :diameter]
    dist_from_center = np.sqrt((x - radius) ** 2 + (y - radius) ** 2)
    return dist_from_center <= radius


def _make_circular_kernel(kernel: NDArray[T], radius: int) -> NDArray[T]:
    kernel[~_make_circular_mask(radius)] = 0
    return kernel


def gaussian_kernel(
    bw: float, radius: int, *, dtype: DTypeLike = np.float32, circular: bool = False
) -> NDArray:
    """
    Generate a 2D Gaussian kernel array.

    Parameters
    ----------
    bw : float
        Bandwidth of the Gaussian.
    radius : int
        Radius of the kernel. Output size will be :math:`2*radius+1`.
    dtype : numpy.typing.DTypeLike
        Datatype of the kernel.
    circular : bool, optional
        Whether to make kernel circular. Values outside `radius` will be set to 0.

    Returns
    -------
    numpy.ndarray
    """
    mask_size = 2 * radius + 1

    dirac = signal.unit_impulse((mask_size, mask_size), idx="mid")

    gaussian_kernel = ndimage.gaussian_filter(
        dirac, bw, output=np.float64, radius=radius
    ).astype(dtype)

    if circular:
        gaussian_kernel = _make_circular_kernel(gaussian_kernel, radius)

    return gaussian_kernel


def epanechnikov_kernel(bw: float, *, dtype: DTypeLike = np.float32) -> np.ndarray:
    """
    Generate a 2D Epanechnikov kernel array.

    :math:`K(x) = 1/2 * c_d^{-1}*(d+2)(1-||x||^2)` if :math:`||x|| < 1` else 0,
    where :math:`d` is the number of dimensions
    and :math:`c_d` the volume of the unit `d`-dimensional sphere.

    Parameters
    ----------
    bw : float
        Bandwidth of the kernel.
    dtype : numpy.typing.DTypeLike, optional
        Datatype of the kernel.

    Returns
    -------
    numpy.ndarray
    """
    # https://doi.org/10.1109/CVPR.2000.854761
    # c_d = pi for d=2

    r = math.ceil(bw)
    dia = 2 * r - 1  # values at r are zero anyways so the kernel matrix can be smaller

    # 1/2 * pi^-1 * (d+2)
    scale = 2 / math.pi

    kernel = np.zeros((dia, dia), dtype=dtype)
    for i in range(dia):
        for j in range(dia):
            x = i - r + 1
            y = j - r + 1
            norm = (x / bw) ** 2 + (y / bw) ** 2
            if norm < 1:
                kernel[i, j] = scale * (1 - norm)

    return kernel / np.sum(kernel)



import os
from typing import NoReturn

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from ._utils_rust import coordinate_as_string


def _get_n_cpus() -> int:
    return len(os.sched_getaffinity(0))


def _get_coordinate_index(
    x: NDArray[np.integer],
    y: NDArray[np.integer],
    *,
    name: str | None = None,
    n_threads: int = 1,
) -> pd.Index:
    x_i32: NDArray[np.int32] = x.astype(np.int32, copy=False)
    y_i32: NDArray[np.int32] = y.astype(np.int32, copy=False)

    return pd.Index(
        coordinate_as_string(x_i32, y_i32, n_threads=n_threads), dtype=str, name=name
    )


def _bin_coordinates(df: pd.DataFrame, bin_size: float) -> pd.DataFrame:
    df = df.assign(
        x=lambda df: _get_bin_coordinate(df["x"].to_numpy(), bin_size),
        y=lambda df: _get_bin_coordinate(df["y"].to_numpy(), bin_size),
    )
    return df


def _get_bin_coordinate(coor: NDArray[np.number], bin_size: float) -> NDArray[np.int32]:
    return np.floor(coor / bin_size).astype(np.int32, copy=False)


def _raise_module_load_error(e: Exception, fn: str, pkg: str, extra: str) -> NoReturn:
    raise ModuleNotFoundError(
        f"`{fn}` requires '{pkg}' to be installed, e.g. via the '{extra}' extra."
    ) from e



import os
from typing import TypeAlias

import numpy as np
from numpy.typing import NDArray
from scipy.sparse import csc_array, csc_matrix, csr_array, csr_matrix

_PathLike: TypeAlias = os.PathLike[str] | str

_Csr: TypeAlias = csr_array | csr_matrix
_Csc: TypeAlias = csc_array | csc_matrix
_Csx: TypeAlias = _Csr | _Csc
_CsxArray: TypeAlias = csc_array | csr_array

_RangeTuple: TypeAlias = tuple[int, int]
_RangeTuple2D: TypeAlias = tuple[_RangeTuple, _RangeTuple]

_Local_Max: TypeAlias = tuple[NDArray[np.int_], NDArray[np.int_]]

_Color: TypeAlias = str | tuple[float, float, float]
_Cmap: TypeAlias = str | list[_Color] | dict[str, _Color]



from typing import TYPE_CHECKING

import pandas as pd

from . import __version__
from ._utils import _raise_module_load_error

if TYPE_CHECKING:
    from pooch import Pooch

# version tags have the format vX.Y.Z but __version__ is X.Y.Z
# we need to modify it otherwise the url is incorrect
version = "v" + __version__


def _get_signature_pooch() -> "Pooch":
    # use indirection to enable pooch as optional dependency w/o lazy loading
    try:
        import pooch

        SIGNATURES = pooch.create(
            path=pooch.os_cache("sainsc"),
            base_url="https://github.com/HiDiHLabs/sainsc/raw/{version}/data/",
            version=version,
            version_dev="main",
            registry={
                "signatures_brain.tsv": "sha256:1e7e3e959ea0a0efdfb8bff2ef1de368757a26f317088122a14dd0b141f7149e",
                "signatures_embryo.tsv": "sha256:8554959144fa2e38b2ce0fa59bd4911d09ef84e0565917371bc7a16e062ad68a",
                "embryo_selection.png": "sha256:431011086d5c537c691d2598611d5ea048c0d0808186ba3f924cd7deb625ef56",
            },
        )
        return SIGNATURES
    except ModuleNotFoundError as e:
        _raise_module_load_error(e, "fetch_*_signatures", pkg="pooch", extra="example")


def fetch_brain_signatures() -> pd.DataFrame:
    """
    Load the Stereo-seq mouse hemibrain cell-type signatures.
    """
    fname = _get_signature_pooch().fetch("signatures_brain.tsv")
    data = pd.read_table(fname, index_col=0)
    return data


def fetch_embryo_signatures() -> pd.DataFrame:
    """
    Load the Stereo-seq embryo cell-type signatures.
    """
    fname = _get_signature_pooch().fetch("signatures_embryo.tsv")
    data = pd.read_table(fname, index_col=0)
    return data


def fetch_embryo_mask() -> str:
    """
    Load the Stereo-seq embryo selection mask image.
    """
    fname = _get_signature_pooch().fetch("embryo_selection.png")
    return fname
